''***************************************
''*  Copyright (c) 2014 Dave Hein       *
''*  See end of file for terms of use.  *
''***************************************

CON

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000

  tiles    = vga#xtiles * vga#ytiles
  tiles32  = tiles * 32

OBJ
  vga : "vga_512x384_bitmap"
  ser : "FullDuplexSerial"

VAR
  long  sync, pixels[tiles32]
  word  colors[tiles]
  long rotmatrix[9]
  long stack[100]

PUB start | i, val
  'waitcnt(clkfreq*3+cnt)
  'ser.start(31, 30, 0, 115200)
  ser.start(7, 6, 0, 115200)
  
  'start vga
  vga.start(16, @colors, @pixels, @sync)

  ' start drawing engine
  _pixels := @pixels
  cognew(@draweng, @command)

  ' start receiving cog
  cognew(Receive, @stack)

  'init colors to cyan on black
  repeat i from 0 to tiles - 1
    colors[i] := $3C00

  repeat
    waitcnt(clkfreq/10+cnt)
    longfill(@pixels, 0, tiles32)
    DrawLineMark(256, 192, 256 + ((190*rotmatrix[1]) ~> 16), 192 - ((190*rotmatrix[2]) ~> 16), 0)
    DrawLineMark(256, 192, 256 + ((190*rotmatrix[4]) ~> 16), 192 - ((190*rotmatrix[5]) ~> 16), 1)
    DrawLineMark(256, 192, 256 + ((190*rotmatrix[7]) ~> 16), 192 - ((190*rotmatrix[8]) ~> 16), 2)

PUB Receive | index, val
  index := 9
  repeat
    val := ser.rx
    if index < 9
      rotmatrix[index++] := (val << 24) ~> 15
    elseif val == 13
      index := 0

PUB PlotPixel(x, y) | index
  if x < 0 or x > 511
    return
  if y < 0 or y > 383
    return
  index := (y << 9) + x
  pixels[index >> 5] |= (1 << (index & 31))

PUB DrawLineMark(x0, y0, x1, y1, mark) | i, j
  DrawLine(x0, y0, x1, y1)
  repeat i from 0 to mark
    j := (i + 1) * 3
    DrawLine(x1 + j, y1-3, x1 + j, y1+3)

PUB DrawLine(x0, y0, x1, y1) | x, y, dx, dy ', m, msum
  dx := x1 - x0
  dy := y1 - y0
  if not dx and not dy
    PlotPixel(x0, y0)
    return
  'msum := 2048
  if || dy < || dx
    m := (dy << 12)/dx
    if x1 < x0
      ||dx
      msum := (y1 << 12) + 2048
      x := x1
    else
      msum := (y0 << 12) + 2048
      x := x0
    count := dx
    xy := x
    command := 1
  else
    m := (dx << 12)/dy
    if y1 < y0
      ||dy
      msum := (x1 << 12) + 2048
      y := y1
    else
      msum := (x0 << 12) + 2048
      y := y0
    count := dy
    xy := y
    command := 2
  repeat while command

' PASM Line Drawing Engine
DAT
command       long      0       ' 1 - Step in X direction, 2 - Step in Y direction
count         long      0       ' Number of steps
msum          long      0       ' Initial Y or X value scaled by 4096 plus 2048 for rounding
m             long      0       ' Slope scaled by 4096
xy            long      0       ' Initial X or Y value depending on command

              org       0

              'Get command, and then get parms if non-zero
draweng       rdlong    _command, par           wz
        if_z  jmp       #draweng
              mov       _addr, par
              add       _addr, #4
              rdlong    _count, _addr
              add       _addr, #4
              rdlong    _msum, _addr
              add       _addr, #4
              rdlong    _m, _addr
              add       _addr, #4
              cmp       _command, #1             wz
        if_z  jmp       #stepx

              ' Step by one along Y axis        
stepy         rdlong    _y, _addr
:loop         mov       _x, _msum
              sar       _x, #12
              call      #_plotpixel
              add       _y, #1
              add       _msum, _m
              djnz      _count, #:loop
              jmp       #draweng_done

              ' Step by one along X axis              
stepx         rdlong    _x, _addr
:loop         mov       _y, _msum
              sar       _y, #12
              call      #_plotpixel
              add       _x, #1
              add       _msum, _m
              djnz      _count, #:loop

              ' Signal completer by writing zero to command              
draweng_done  wrlong    _zero, par
              jmp       #draweng

              ' Plot a pixel at X, Y              
_plotpixel
              'jmp       _plotpixel_ret
              cmps      _x, #0                  wc
        if_c  jmp       _plotpixel_ret
              cmps      _x, _k512               wc
        if_nc jmp       _plotpixel_ret
              cmps      _y, #0                  wc
        if_c  jmp       _plotpixel_ret
              cmp       _y, #384                wc
        if_nc jmp       _plotpixel_ret

              ' index := (y << 9) + x
              mov       _index, _y
              shl       _index, #9
              add       _index, _x
              ' pixels[index >> 5] |= (1 << (index & 31))
              mov       _lx, _index
              shr       _lx, #5
              mov       _bx, _index
              and       _bx, #31
              mov       _addr, _pixels
              shl       _lx, #2
              add       _addr, _lx
              rdlong    _temp1, _addr
              mov       _temp2, #1
              shl       _temp2, _bx
              or        _temp1, _temp2
              wrlong    _temp1, _addr
_plotpixel_ret ret

_pixels       long      0
_zero         long      0
_k512         long      512
_addr         res       1
_temp1        res       1
_temp2        res       1
_command      res       1
_index        res       1
_x            res       1
_y            res       1
_msum         res       1
_m            res       1
_lx           res       1
_bx           res       1
_count        res       1 